// Name: Jonathan Turner
// Class: CS 4306/01
// Term: Spring 2024
// Instructor: Dr. Haddad
// Assignment: 2
// IDE Name: IntelliJ
/*
Algorithm Design Block
    Algorithm title: Finding Equal Disjoint Sets

    Logical steps:
        Step 1: Get the number of items in the set as user input.
        Step 2: Get the elements in the set as user input.
        Step 3: Add all the elements to the list and save it to a variable.
        Step 4: If the sum is odd, there is no disjoint sets and it is not possible.
        Step 5: Set a variable that will hold the first set and second set representing the disjoint and initialize it to null.
        Step 6: Create a List that holds a list of elements which represent each subset.
        Step 7: Generate the subsets using a subset generation method. (Ex. Bit Shift Counting)
        Step 8: Iterate through all the subsets that were generate.
        Step 9: If the current iteration's sum is equal to half of the total sum, move to step 11. Otherwise, continue iterating.
        Step 10: If the value is not found, then there is no disjoint equal subsets. End algorithm.
        Step 11: Assign the first set's variable to the current iteration's values.
        Step 12: Assign the second set's variable to the values of the original set that are not in the first set.
        Step 13: Return the results of the two equal disjoint sets that were found.

    Pseudocode Syntax:
        count <- input
        elements <- input
        first_set, second_set <- empty

        sum <- 0
        for each number in elements:
            sum <- sum + number

        if sum % 2 is 0:
            subsets <- generate_subsets(elements)
            for each subset in subsets:
                current_sum <- 0
                for each value in subset:
                    current_sum <- current_sum + value
                if sum / 2 is current_sum:
                    first_set <- subset

            if first_set is not empty:
                second_set <- elements - first_set
            else
                return "no equal disjoint subsets"

        return first_set, second_set

    Big-O Analysis: (Based on Implementation)

    n = the number of elements in the set.
    k = the number of elements in the current subset.

    |  init Comps  |  init values  |  i  | totalSum | generateSubsets | subset |    value    |      subsetTotal    | if(half) | asignSet | secondSubset  | printResults
    -------------------------------------------------------------------------------------------min-------max-------------------------min--max---------------------------------
    |      1       |      1        |  1  |    1     |     2^n         |   1    |   2^(n)+1   | 1        2^(n)+1    |   2^(n)  | 0    1   | 1             | 1
    |              |               |  2  |    1     |                 |   2    |   2^(n)+2   | 1        2^(n)+2    |   2^(n)  | 0    1   | 2             | 1
    |              |               |  3  |    1     |                 |   3    |   2^(n)+3   | 1        2^(n)+3    |   2^(n)  | 0    1   | 3             | 1
    |              |               |  4  |    1     |                 |   4    |   2^(n)+4   | 1        2^(n)+4    |   2^(n)  | 0    1   | 4             | 1
    |              |               | ... |    ...   |                 |  ...   |     ...     | 1          ...      |    ...   |...  ...  |...            | 1
    |              |               |  n  |    1     |                 |  2^n   |   2^(n)+k   | 1        2^(n)+k    |   2^(n)  | 0    1   | k             | 1
    --------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1            | 1             |  X  | n        | 2^n             | X      | X           | 1        k*2^(n) +  | n*2^(n)  | 0    2^n | k             | 1
                                                                                                      (k^(2)+k)/2 |
    Best-Case Scenario: 1 + 1 + n + 2^n + 1 + (n*2^(n)) + 0 + k + 1 = n + 2^n + (n*2^(n)) + k + 4
    Worse-Case Scenario: 1 + 1 + n + 2^n + k*2^n+(k*2^(n))/2 + n*2^(n) + 2^n + k + 1 = 2^n + k*2^n + (k*2^(n))/2 + n*2^(n) + k + 4
    Average-Case: (n*2^(n) + 2^n + k + 4 + 2^n + k*2^n + (k*2^(n))/2 + n*2^(n) + k + 4)/2

    Big O => O(2^n)
*/
package Assignments.A2;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Scanner;

/**
 * Generates, if possible, two equal subsets of a given set of elements that are equal to each other.
 *
 * @author Jonathan Turner
 * @version Spring 2024
 */
public class Partition {

    private int[] set;
    private List<List<Integer>> subsets;

    public static void main(String[] args) {
        Partition program = new Partition();
        program.start();
    }

    /**
     * Used as the main sentential loop and loops until the exit option is selected.
     *
     * @precondition none
     * @postcondition none
     */
    public void start() {
        int option = this.getOption();
        while (option != 4) {
            this.executeOption(option);
            option = this.getOption();
        }
    }

    /**
     * Performs the action related to the option that was chosen.
     *
     * @precondition none
     * @postcondition the action was performed.
     *
     * @param option the option that was chosen.
     */
    private void executeOption(int option) {
        if (option == 1) {
            System.out.println("Please specify the size of the set.");
            System.out.print("Size: ");
            int newSize = this.getIntegerInput();
            while (newSize < 1) {
                System.out.println("Please enter a valid set size.");
                System.out.print("Size: ");
                newSize = this.getIntegerInput();
            }
            this.set = new int[newSize];
        } if (option == 2) {
            if (this.set == null) {
                System.out.println("Please enter a set size and values ");
                System.out.println();
                return;
            }
            this.set = this.getValues();
        } else {
            if (this.set == null || this.set[0] == 0) {
                System.out.println("Please enter a set size and values before executing the algorithm.");
                System.out.println();
                return;
            }
            this.performAlgorithm();
        }
        System.out.println();
    }

    /**
     * Prints out the menu of options, asks for an input, and if that input is invalid it prints an error and
     * prompts the user again for the input.
     *
     * @precondition none
     * @postcondition none
     *
     * @return the option selected
     */
    private int getOption() {
        this.displayMenu(); // Prints out the menu

        // Asks for the option and begins the loop until a valid option is gathered.
        System.out.print("Enter option number: ");
        int input = getIntegerInput();
        while (input < 1 || input > 4) { // Compares it to the valid options available.
            System.out.println("\nPlease enter a valid input.");
            System.out.print("Enter option number: ");
            input = getIntegerInput();
        }
        System.out.println();
        return input;
    }

    /**
     * Displays the menu options.
     *
     * @precondition none
     * @postcondition the menu is displayed.
     */
    private void displayMenu() {
        System.out.println("-----------------MAIN MENU--------------");
        System.out.println("1. Read set size (number of integers)");
        System.out.println("2. Read set elements (integer values)");
        System.out.println("3. Run algorithm and display outputs");
        System.out.println("4. Exit program");
        System.out.println();
    }

    /**
     * Gets user input to a number. A prompt must be provided prior to running this method.
     * If the input was not an integer/(unable to be autoboxed), returns -1.
     *
     * @precondition none
     * @postcondition none
     *
     * @return the integer input, if invalid -1.
     */
    private int getIntegerInput() {
        Scanner sc = new Scanner(System.in);
        try {
            String textInput = sc.nextLine();
            return Integer.parseInt(textInput);
        } catch (Exception e) {
            return -1;
        }
    }

    /**
     * Gets the values for the set. If the input is invalid, it will prompt the user again.
     *
     * @precondition none
     * @postcondition the values are gathered.
     *
     * @return the values of the set.
     */
    private int[] getValues() {
        Scanner sc = new Scanner(System.in);
        String input = null;
        int[] values = new int[this.set.length];
        System.out.println("Please enter a list of " + this.set.length + " positive values. (Ex. 34 35 21 23)");
        while (values[0] == 0) {
            System.out.print("Values: ");
            input = sc.nextLine();
            String[] stringRep = input.split(" ");

            /* If the input values are not the same number as the set size */
            if (stringRep.length != values.length || input.isEmpty() ) {
                System.out.println("Please enter a valid list of " + this.set.length + " positive values. (Ex. 34 35 21 23)");
                continue;
            }

            /* Attempts to convert each value from String to Integer. */
            try {
                for (int i = 0; i < values.length; i++) {
                    values[i] = Integer.parseInt(stringRep[i]);
                    if (values[i] < 1) {
                        System.out.println("Please enter a valid list of " + this.set.length + " positive values. (Ex. 34 35 21 23)");
                        Arrays.fill(values, 0);
                        break;
                    }
                }
            } catch (Exception e) {
                System.out.println("Please enter a valid list of " + this.set.length + " positive values. (Ex. 34 35 21 23)");
            }
        }
        return values;
    }

    /**
     * Iterates through all the possible subsets generated by this.generateSubsets
     * and if a subset is found that is exactly half the size it sets the firstset
     * to those values and the second subset to the missing values.
     *
     * @precondition set.length != 0
     * @postcondition the resulting sets are printed.
     */
    public void performAlgorithm() {
        int totalSum = 0;
        for (int i = 0; i < this.set.length; i++) {
            totalSum += this.set[i];
        }
        List<Integer> firstSubset = null;
        List<Integer> secondSubset = null;

        /* If the total is odd, there cannot be equal disjoint subsets. */
        if (totalSum % 2 == 1) {
            printResults(firstSubset, secondSubset);
            return;
        }

        /*
         * Checks through every subset that was generated,
         * if a subset is found it sets the firstSubset to the found subset.
        */
        generateSubsets();
        for (List<Integer> subset : subsets) {
            int subsetTotal = 0;
            for (int value : subset) {
                subsetTotal += value;
            }
            if (totalSum / 2 == subsetTotal) {
                firstSubset = subset;
                break;
            }
        }

        /* Checks if a list was found, if so creates second sublist. */
        if (firstSubset != null) {
            secondSubset = new ArrayList<>();
            for (int secondValue : this.set) {
                if (!firstSubset.contains(secondValue)) {
                    secondSubset.add(secondValue);
                }
            }
        }
        printResults(firstSubset, secondSubset);
    }

    /**
     * Prints the results of the algorithm based on the two subsets. Prints the setsize and the original set.
     * If the subsets were null (No match found) print output associated, otherwise print the both lists.
     *
     * @precondition none
     * @postcondition the output is displayed.
     *
     * @param firstSubset the first subset.
     * @param secondSubset the second subset.
     */
    private void printResults(List<Integer> firstSubset, List<Integer> secondSubset) {
        int setSize = this.set.length;
        String original = "";
        for (int values : this.set) {
            original += values + " ";
        }
        String first = "{";
        String second = "{";
        if (firstSubset != null) {
            for (int fValues : firstSubset) {
                first += fValues + ",";
            }
            first = first.substring(0, first.length()-1) + "}";
            for (int sValues : secondSubset) {
                second += sValues + ",";
            }
            second = second.substring(0, second.length()-1) + "}";
        } else {
            first = "No disjoint subsets with the same sum";
            second = "of their elements found";
        }

        System.out.println("Set size:               " + setSize + " integers");
        System.out.println("Integer values:         " + original);
        System.out.println("Disjoint subsets with same sum: " + first);
        System.out.println("                                " + second);
    }

    /**
     * Generates the list of subsets using Bit Shift Counting by iterating over all combinations of 1 to n.
     * After shifting the counter, the current subset is created initialized and if 1 shifted by j is logically
     * ANDed with the i and is not equal to 0, it adds the value j represents to the subset.
     *
     * @precondition none
     * @postcondition the subsets are generated and added to the field.
     */
    public void generateSubsets() {
        List<List<Integer>> subsets = new ArrayList<>();
        int n = this.set.length;

        for (int i = 0; i < (1 << n); i++) {
            List<Integer> currentSubset = new ArrayList<>();
            for (int j = 0; j < n; j++) {
                if ((i & (1 << j)) != 0) {
                    currentSubset.add(this.set[j]);
                }
            }
            subsets.add(currentSubset);
        }
        this.subsets = subsets;
    }
}
